/********************************COPYRIGHT*************************************

    Copyright (C) 2002 - 2005 Brendt Wohlberg  <software@wohlberg.net>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

********************************COPYRIGHT*************************************/

/*
  TIFF directory access functions.

  The approach used here is based closely on that of the "tiffdump"
  utility distributed with libtiff.

  Most recent modification: 19 February 2005
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tiffdir.h" 


/* Endian correction on main part of directory entry */
void InitialDirEntrySwab(TIFFDirEntry* tdep, int swabflag) {
  if (swabflag) {
      TIFFSwabShort(&tdep->tdir_tag);
      TIFFSwabShort(&tdep->tdir_type);
      TIFFSwabLong(&tdep->tdir_count);
      TIFFSwabLong(&tdep->tdir_offset);
    }
}


/* Get size in bytes of a directory entry value */
int DirEntrySize(TIFFDirEntry* tdep) {
  /* Array of sizes corresponding to each value in TIFFDataType */
  const static int size[] = {0, 1, 1, 2, 4, 8, 1, 1, 2, 4, 8, 4, 8};
  const int nsz = 13;
 
  /* Get number of bytes required for the value */
  if (tdep->tdir_type < nsz)
    return size[tdep->tdir_type]*tdep->tdir_count;
  else
    return 0;
}


/* Get directory entry value */
void* DirEntryValue(TIFFDirEntry* tdep, FILE* fp, int swabflag) {
  int nbytes = 0;
  void* valp = 0;

  nbytes = DirEntrySize(tdep);
  if (nbytes == 0)
    return 0;

  /* Allocate memory for the value */
  if (!(valp = malloc(nbytes)))
    return 0;
  if (nbytes <= 4) {
    /* If the value requires 4 or fewer bytes, it's in the offset */
    memcpy(valp, &tdep->tdir_offset, nbytes);
  } else {
    /* If the value requires more than 4 bytes, seek to the offset
       and read it */
    if (fseek(fp, tdep->tdir_offset, SEEK_SET) == -1) {
      free(valp);
      return 0;
    } else {
      if (fread(valp, 1, nbytes, fp) != nbytes) {
	free(valp);
	return 0;
      }
    }

    /* Endian correction to value obtained via offset indirection
       (not necessary for value in offset since that's corrected in
       the call to InitialDirEntrySwab */
    if (swabflag) {
      switch (tdep->tdir_type) {
      case TIFF_SHORT:
      case TIFF_SSHORT:
	TIFFSwabArrayOfShort((uint16*) valp, tdep->tdir_count);
	break;
      case TIFF_LONG:
      case TIFF_SLONG:
      case TIFF_FLOAT:
	TIFFSwabArrayOfLong((uint32*) valp, tdep->tdir_count);
	break;
      case TIFF_RATIONAL:
      case TIFF_SRATIONAL:
	TIFFSwabArrayOfLong((uint32*) valp, 2*tdep->tdir_count);
	break;
      case TIFF_DOUBLE:
	TIFFSwabArrayOfDouble((double*) valp, tdep->tdir_count);
	break;
      }
    }

  }

  return valp;
}


/* Open a TIFF IFD */
TIFFDirEntry* OpenDirectory(FILE* fp, uint32 diroff, uint16* ndep, 
			   int swabflag) {
  TIFFDirEntry* tdep;
  int k;
  
  /* Seek to position of directory offset */
  if (fseek(fp, diroff, SEEK_SET) == -1)
     return 0;

  /* Read number of directory entries */
  if (fread(ndep, sizeof(*ndep), 1, fp) != 1)
    return 0;
  
  /* Endian correction if necessary */
  if (swabflag)
    TIFFSwabShort(ndep);
  
  /* Allocate memory for directory entries and read them */
  if ((tdep = malloc((*ndep)*sizeof(TIFFDirEntry))) == 0)
    return 0;
  if (fread(tdep, sizeof(TIFFDirEntry), *ndep, fp) != (*ndep)) 
    return 0;

  /* Endian correction to main part of each directory entry */
  for (k = 0; k < *ndep; k++)
    InitialDirEntrySwab(&tdep[k], swabflag);

  return tdep;
}



/* Clean up a TIFF IFD structure */
void CloseDirectory(TIFFDirEntry* tdep) {
  free(tdep);
}



/* Get a TIFF IFD entry */
TIFFDirEntry* TagDirEntry(TIFFDirEntry* tdep, uint16 nde, uint16 tag) {
  uint16 k;

  /* Work through each directory entry */
  for (k = 0; k < nde; k++) {
    if (tdep[k].tdir_tag == tag) {
      return &tdep[k];
      break;
    }
}
  return 0;
}


/* Get a tag value within a TIFF IFD */
void* TagValue(TIFFDirEntry* tdep, uint16 nde, FILE* fp, uint16 tag,
	       int swabflag) {
  uint16 k;
  void* vp = 0;

  /* Work through each directory entry */
  for (k = 0; k < nde; k++) {
    if (tdep[k].tdir_tag == tag) {
      if ((vp = DirEntryValue(&tdep[k],fp,swabflag)) == 0)
	return 0;
      break;
    }
  }
  return vp;
}



/* Open a Nikon Scan format MakerNote IFD */
TIFFDirEntry* OpenMakerNote(FILE* fp, uint32 diroff, uint16* ndep, 
			    uint32* thop, int swabflag) {
  TIFFDirEntry* tdep;
  char makerid[6];
  uint32 u;
  uint16 magic;
  uint16 id;
  uint32 off;
  int k;
  
  /* Zero the MakerNote TIFF header offset pointer */
  *thop = 0;

  /* Seek to position of directory offset */
  if (fseek(fp, diroff, SEEK_SET) == -1)
     return 0;
  /* Read first 6 bytes to allow check for MakerNote in "TIFF 
     header" scheme */
  if (fread(makerid, 1, 6, fp) != 6)
    return 0;

  /* Check for MakerNote in "TIFF header" scheme */
  if (strcmp(makerid, "Nikon") != 0) {
    /* MakerNote is in standard IFD format */
    return OpenDirectory(fp, diroff, ndep, swabflag); 
  } else {
    /* MakerNote uses "TIFF header" scheme */

    /* Read apparently unused uint32 */
    if (fread(&u, sizeof(u), 1, fp) != 1)
    return 0;

    /* Record start of MakerNote TIFF header */
    *thop = ftell(fp);

    /* Read endian indicator magic number */
    if (fread(&magic, sizeof(magic), 1, fp) != 1)
    return 0;
    /* Ensure magic number is valid */
    if (!(magic == TIFF_BIGENDIAN || magic == TIFF_LITTLEENDIAN))
      return 0;
    /* Set endian correction flag based on magic number */
    swabflag = REQUIRE_ENDIAN_CORRECTION(magic);
    /* Read secondary magic number */
    if (fread(&id, sizeof(id), 1, fp) != 1)
      return 0;
    /* Perform endian correction if necessary */
    if (swabflag)
     TIFFSwabShort(&id);
   /* Ensure secondary magic number is valid */
    if (id != 42)
      return 0;
    /* Read offset to start of IFD content */
    if (fread(&off, sizeof(off), 1, fp) != 1)
      return 0;
    /* Seek to offset */
    if (fseek(fp, off-8, SEEK_CUR) == -1)
     return 0;

    /* Read number of directory entries */
    if (fread(ndep, sizeof(*ndep), 1, fp) != 1)
      return 0;
  
    /* Perform endian correction if necessary */
    if (swabflag)
      TIFFSwabShort(ndep);
  
    /* Allocate memory for directory entries and read them */
    if ((tdep = malloc((*ndep)*sizeof(TIFFDirEntry))) == 0)
      return 0;
    if (fread(tdep, sizeof(TIFFDirEntry), *ndep, fp) != (*ndep)) 
      return 0;
    
    for (k = 0; k < *ndep; k++) {
      /* Endian correction to main part of each directory entry */
      InitialDirEntrySwab(&tdep[k], swabflag);
      /* Change offsets relative to TIFF header to file offsets */
      if (DirEntrySize(&tdep[k]) > 4)
	tdep[k].tdir_offset += *thop;
    }

    return tdep;
  }
}

